@@ -74,6 +74,9 @@ class Feature:
 
     @classmethod
     def from_response_json(cls, json_dict):
+        if json_dict is None:
+            return None
+
         json_decamelized = humps.decamelize(json_dict)
         return cls(**json_decamelized)
 
@@ -169,6 +172,9 @@ class Feature:
     def __gt__(self, other):
         return filter.Filter(self, filter.Filter.GT, other)
 
+    def contains(self, other):
+        return filter.Filter(self, filter.Filter.IN, json.dumps(other))
+
     def __str__(self):
         return self.json()
 
