@@ -22,10 +22,11 @@ import numpy as np
 import avro.schema
 from typing import Optional, Union, Any, Dict, List, TypeVar
 
-from hsfs import util, engine, feature, storage_connector as sc
+from hsfs import util, engine, feature, user, storage_connector as sc
 from hsfs.core import (
     feature_group_engine,
     statistics_engine,
+    code_engine,
     data_validation_engine,
     feature_group_base_engine,
     on_demand_feature_group_engine,
@@ -37,13 +38,15 @@ from hsfs.client.exceptions import FeatureStoreException
 
 
 class FeatureGroupBase:
-    def __init__(self, featurestore_id):
+    def __init__(self, featurestore_id, validation_type):
+        self._validation_type = validation_type.upper()
         self._feature_group_base_engine = (
             feature_group_base_engine.FeatureGroupBaseEngine(featurestore_id)
         )
         self._statistics_engine = statistics_engine.StatisticsEngine(
             featurestore_id, self.ENTITY_TYPE
         )
+        self._code_engine = code_engine.CodeEngine(featurestore_id, self.ENTITY_TYPE)
         self._expectations_api = expectations_api.ExpectationsApi(
             featurestore_id, "featuregroups"
         )
@@ -258,7 +261,7 @@ class FeatureGroupBase:
         return self
 
     def update_description(self, description: str):
-        """Update the description of the feature gorup.
+        """Update the description of the feature group.
 
         # Arguments
             description: str. New description string.
@@ -270,7 +273,8 @@ class FeatureGroupBase:
         return self
 
     def attach_expectation(self, expectation):
-        """Get feature group expectations. Gets all expectations if no expectation name is specified.
+        """Attach a feature group expectation. If feature group validation is not already enabled, it will be enabled
+        and set to the stricter setting.
 
         # Arguments
             name: The expectation name.
@@ -279,10 +283,14 @@ class FeatureGroupBase:
             `Expectation`. The expectation metadata object.
 
         """
+        # Turn on validation for this FG and set stricter setting
+        if self._validation_type == "NONE":
+            self._validation_type = "STRICT"
+
         return self._expectations_api.attach(self, expectation.name)
 
     def detach_expectation(self, expectation):
-        """Get feature group expectations. Gets all expectations if no expectation name is specified.
+        """Remove an expectation from a feature group.
 
         # Arguments
             name: The expectation name.
@@ -318,7 +326,7 @@ class FeatureGroupBase:
         return self._expectations_api.get(name, self)
 
     def get_validations(self, validation_time=None, commit_time=None):
-        """Get feature group data validation results based on the attached expectations
+        """Get feature group data validation results based on the attached expectations.
 
         # Arguments
            validation_time: The data validation time, when the data validation started.
@@ -348,7 +356,7 @@ class FeatureGroupBase:
                 f"Expected type `str`, got `{type(name)}`. "
                 "Features are accessible by name."
             )
-        feature = [f for f in self._features if f.name == name]
+        feature = [f for f in self.__getattribute__("_features") if f.name == name]
         if len(feature) == 1:
             return feature[0]
         else:
@@ -382,6 +390,15 @@ class FeatureGroupBase:
         """Get the latest computed statistics for the feature group."""
         return self._statistics_engine.get_last(self)
 
+    @property
+    def primary_key(self):
+        """List of features building the primary key."""
+        return self._primary_key
+
+    @primary_key.setter
+    def primary_key(self, new_primary_key):
+        self._primary_key = [pk.lower() for pk in new_primary_key]
+
     def get_statistics(self, commit_time: str = None):
         """Returns the statistics for this feature group at a specific time.
 
@@ -425,6 +442,15 @@ class FeatureGroupBase:
                 util.StorageWarning,
             )
 
+    @property
+    def event_time(self):
+        """Event time feature in the feature group."""
+        return self._event_time
+
+    @event_time.setter
+    def event_time(self, feature_name):
+        self._event_time = feature_name
+
 
 class FeatureGroup(FeatureGroupBase):
     CACHED_FEATURE_GROUP = "CACHED_FEATURE_GROUP"
@@ -451,14 +477,15 @@ class FeatureGroup(FeatureGroupBase):
         validation_type="NONE",
         expectations=None,
         online_topic_name=None,
+        event_time=None,
     ):
-        super().__init__(featurestore_id)
+        super().__init__(featurestore_id, validation_type)
 
         self._feature_store_id = featurestore_id
         self._feature_store_name = featurestore_name
         self._description = description
         self._created = created
-        self._creator = creator
+        self._creator = user.User.from_response_json(creator)
         self._version = version
         self._name = name
         self._id = id
@@ -475,10 +502,11 @@ class FeatureGroup(FeatureGroupBase):
 
         self._avro_schema = None
         self._online_topic_name = online_topic_name
+        self._event_time = event_time
 
-        if id is not None:
+        if self._id:
             # initialized by backend
-            self._primary_key = [
+            self.primary_key = [
                 feat.name for feat in self._features if feat.primary is True
             ]
             self._partition_key = [
@@ -509,7 +537,6 @@ class FeatureGroup(FeatureGroupBase):
             )
             self.statistics_config = statistics_config
 
-        self._validation_type = validation_type.upper()
         if expectations is not None:
             self._expectations_names = [
                 expectation.name for expectation in expectations
@@ -690,6 +717,8 @@ class FeatureGroup(FeatureGroupBase):
                 * key `wait_for_job` and value `True` or `False` to configure
                   whether or not to the save call should return only
                   after the Hopsworks Job has finished. By default it waits.
+                * key `mode` instruct the ingestion job on how to deal with corrupted
+                  data. Values are PERMISSIVE, DROPMALFORMED or FAILFAST. Default FAILFAST.
 
 
         # Returns
@@ -705,6 +734,7 @@ class FeatureGroup(FeatureGroupBase):
 
         # fg_job is used only if the hive engine is used
         fg_job = self._feature_group_engine.save(self, feature_dataframe, write_options)
+        self._code_engine.save_code(self)
         if self.statistics_config.enabled and engine.get_type() == "spark":
             # Only compute statistics if the engine is Spark.
             # For Hive engine, the computation happens in the Hopsworks application
@@ -776,6 +806,8 @@ class FeatureGroup(FeatureGroupBase):
                 * key `wait_for_job` and value `True` or `False` to configure
                   whether or not to the insert call should return only
                   after the Hopsworks Job has finished. By default it waits.
+                * key `mode` instruct the ingestion job on how to deal with corrupted
+                  data. Values are PERMISSIVE, DROPMALFORMED or FAILFAST. Default FAILFAST.
 
         # Returns
             `FeatureGroup`. Updated feature group metadata object.
@@ -791,6 +823,7 @@ class FeatureGroup(FeatureGroupBase):
             write_options,
         )
 
+        self._code_engine.save_code(self)
         if engine.get_type() == "spark":
             # Only compute statistics if the engine is Spark,
             # if Hive, the statistics are computed by the application doing the insert
@@ -920,6 +953,22 @@ class FeatureGroup(FeatureGroupBase):
             )
         self._feature_group_engine.commit_delete(self, delete_df, write_options)
 
+    def as_of(self, wallclock_time):
+        """Get Query object to retrieve all features of the group at a point in the past.
+
+        This method selects all features in the feature group and returns a Query object
+        at the specified point in time. This can then either be read into a Dataframe
+        or used further to perform joins or construct a training dataset.
+
+        # Arguments
+            wallclock_time: Datetime string. The String should be formatted in one of the
+                following formats `%Y%m%d`, `%Y%m%d%H`, `%Y%m%d%H%M`, or `%Y%m%d%H%M%S`.
+
+        # Returns
+            `Query`. The query object with the applied time travel condition.
+        """
+        return self.select_all().as_of(wallclock_time)
+
     def update_description(self, description: str):
         """Update the description of the feature gorup.
 
@@ -1036,8 +1085,12 @@ class FeatureGroup(FeatureGroupBase):
     @classmethod
     def from_response_json(cls, json_dict):
         json_decamelized = humps.decamelize(json_dict)
-        _ = json_decamelized.pop("type", None)
-        return cls(**json_decamelized)
+        if isinstance(json_decamelized, dict):
+            _ = json_decamelized.pop("type", None)
+            return cls(**json_decamelized)
+        for fg in json_decamelized:
+            _ = fg.pop("type", None)
+        return [cls(**fg) for fg in json_decamelized]
 
     def update_from_response_json(self, json_dict):
         json_decamelized = humps.decamelize(json_dict)
@@ -1062,6 +1115,7 @@ class FeatureGroup(FeatureGroupBase):
             "statisticsConfig": self._statistics_config,
             "validationType": self._validation_type,
             "expectationsNames": self._expectations_names,
+            "eventTime": self._event_time,
         }
 
     def _get_table_name(self):
@@ -1125,11 +1179,6 @@ class FeatureGroup(FeatureGroupBase):
     def location(self):
         return self._location
 
-    @property
-    def primary_key(self):
-        """List of features building the primary key."""
-        return self._primary_key
-
     @property
     def online_enabled(self):
         """Setting if the feature group is available in online storage."""
@@ -1203,10 +1252,6 @@ class FeatureGroup(FeatureGroupBase):
     def time_travel_format(self, new_time_travel_format):
         self._time_travel_format = new_time_travel_format
 
-    @primary_key.setter
-    def primary_key(self, new_primary_key):
-        self._primary_key = [pk.lower() for pk in new_primary_key]
-
     @partition_key.setter
     def partition_key(self, new_partition_key):
         self._partition_key = [pk.lower() for pk in new_partition_key]
@@ -1246,6 +1291,7 @@ class OnDemandFeatureGroup(FeatureGroupBase):
         name=None,
         version=None,
         description=None,
+        primary_key=None,
         featurestore_id=None,
         featurestore_name=None,
         created=None,
@@ -1253,22 +1299,24 @@ class OnDemandFeatureGroup(FeatureGroupBase):
         id=None,
         features=None,
         statistics_config=None,
+        event_time=None,
         validation_type="NONE",
         expectations=None,
     ):
-        super().__init__(featurestore_id)
+        super().__init__(featurestore_id, validation_type)
 
         self._feature_store_id = featurestore_id
         self._feature_store_name = featurestore_name
         self._description = description
         self._created = created
-        self._creator = creator
+        self._creator = user.User.from_response_json(creator)
         self._version = version
         self._name = name
         self._query = query
         self._data_format = data_format.upper() if data_format else None
         self._path = path
         self._id = id
+        self._event_time = event_time
 
         self._feature_group_engine = (
             on_demand_feature_group_engine.OnDemandFeatureGroupEngine(featurestore_id)
@@ -1281,6 +1329,11 @@ class OnDemandFeatureGroup(FeatureGroupBase):
                 if features
                 else None
             )
+            self.primary_key = (
+                [feat.name for feat in self._features if feat.primary is True]
+                if self._features
+                else []
+            )
             self.statistics_config = statistics_config
 
             self._options = (
@@ -1289,6 +1342,7 @@ class OnDemandFeatureGroup(FeatureGroupBase):
                 else None
             )
         else:
+            self.primary_key = primary_key
             self.statistics_config = statistics_config
             self._features = features
             self._options = options
@@ -1299,7 +1353,7 @@ class OnDemandFeatureGroup(FeatureGroupBase):
             )
         else:
             self._storage_connector = storage_connector
-        self._validation_type = validation_type.upper()
+
         if expectations is not None:
             self._expectations_names = [
                 expectation.name for expectation in expectations
@@ -1309,11 +1363,12 @@ class OnDemandFeatureGroup(FeatureGroupBase):
 
     def save(self):
         self._feature_group_engine.save(self)
+        self._code_engine.save_code(self)
         if self._validation_type != "NONE":
             self.validate()
 
         if self.statistics_config.enabled:
-            self._statistics_engine.compute_statistics(self, self.read)
+            self._statistics_engine.compute_statistics(self, self.read())
 
     def read(self, dataframe_type="default"):
         """Get the feature group as a DataFrame."""
@@ -1347,9 +1402,14 @@ class OnDemandFeatureGroup(FeatureGroupBase):
     @classmethod
     def from_response_json(cls, json_dict):
         json_decamelized = humps.decamelize(json_dict)
-        _ = json_decamelized.pop("online_topic_name", None)
-        _ = json_decamelized.pop("type", None)
-        return cls(**json_decamelized)
+        if isinstance(json_decamelized, dict):
+            _ = json_decamelized.pop("online_topic_name", None)
+            _ = json_decamelized.pop("type", None)
+            return cls(**json_decamelized)
+        for fg in json_decamelized:
+            _ = fg.pop("online_topic_name", None)
+            _ = fg.pop("type", None)
+        return [cls(**fg) for fg in json_decamelized]
 
     def update_from_response_json(self, json_dict):
         json_decamelized = humps.decamelize(json_dict)
@@ -1378,6 +1438,7 @@ class OnDemandFeatureGroup(FeatureGroupBase):
             "storageConnector": self._storage_connector.to_dict(),
             "type": "onDemandFeaturegroupDTO",
             "statisticsConfig": self._statistics_config,
+            "eventTime": self._event_time,
             "validationType": self._validation_type,
             "expectationsNames": self._expectations_names,
         }
