@@ -24,8 +24,9 @@ class Statistics:
     def __init__(
         self,
         commit_time,
-        content,
+        content=None,
         feature_group_commit_id=None,
+        split_statistics=None,
         href=None,
         expand=None,
         items=None,
@@ -34,7 +35,12 @@ class Statistics:
     ):
         self._commit_time = commit_time
         self._feature_group_commit_id = feature_group_commit_id
-        self._content = json.loads(content)
+        # if split_statistics is provided then content will be None
+        if split_statistics is None:
+            self._content = json.loads(content)
+        else:
+            self._content = content
+        self._split_statistics = split_statistics
 
     @classmethod
     def from_response_json(cls, json_dict):
@@ -50,6 +56,7 @@ class Statistics:
             "commitTime": self._commit_time,
             "featureGroupCommitId": self._feature_group_commit_id,
             "content": json.dumps(self._content),
+            "splitStatistics": self._split_statistics,
         }
 
     def json(self):
@@ -66,3 +73,7 @@ class Statistics:
     @property
     def content(self):
         return self._content
+
+    @property
+    def split_statistics(self):
+        return self._split_statistics
