@@ -1,4 +1,3 @@
-#
 #   Copyright 2020 Logical Clocks AB
 #
 #   Licensed under the Apache License, Version 2.0 (the "License");
@@ -30,6 +29,7 @@ from hsfs.core import (
     training_dataset_engine,
     tfdata_engine,
     statistics_engine,
+    code_engine,
     transformation_function_engine,
 )
 from hsfs.constructor import query
@@ -76,7 +76,7 @@ class TrainingDataset:
         self._from_query = from_query
         self._querydto = querydto
         self._feature_store_id = featurestore_id
-        self._prepared_statement_connection = None
+        self._prepared_statement_engine = None
         self._prepared_statements = None
         self._serving_keys = None
         self._transformation_functions = transformation_functions
@@ -93,6 +93,8 @@ class TrainingDataset:
             featurestore_id, self.ENTITY_TYPE
         )
 
+        self._code_engine = code_engine.CodeEngine(featurestore_id, self.ENTITY_TYPE)
+
         self._transformation_function_engine = (
             transformation_function_engine.TransformationFunctionEngine(featurestore_id)
         )
@@ -163,11 +165,15 @@ class TrainingDataset:
         user_version = self._version
         user_stats_config = self._statistics_config
         # td_job is used only if the hive engine is used
-        td_job = self._training_dataset_engine.save(self, features, write_options)
+        training_dataset, td_job = self._training_dataset_engine.save(
+            self, features, write_options
+        )
+        self.storage_connector = training_dataset.storage_connector
         # currently we do not save the training dataset statistics config for training datasets
         self.statistics_config = user_stats_config
+        self._code_engine.save_code(self)
         if self.statistics_config.enabled and engine.get_type() == "spark":
-            self._statistics_engine.compute_statistics(self, self.read())
+            self.compute_statistics()
         if user_version is None:
             warnings.warn(
                 "No version provided for creating training dataset `{}`, incremented version to `{}`.".format(
@@ -224,6 +230,7 @@ class TrainingDataset:
             self, features, write_options, overwrite
         )
 
+        self._code_engine.save_code(self)
         self.compute_statistics()
 
         return td_job
@@ -235,12 +242,18 @@ class TrainingDataset:
 
         # Arguments
             split: Name of the split to read, defaults to `None`, reading the entire
-                training dataset.
+                training dataset. If the training dataset has split, the `split` parameter
+                is mandatory.
             read_options: Additional read options as key/value pairs, defaults to `{}`.
         # Returns
             `DataFrame`: The spark dataframe containing the feature data of the
                 training dataset.
         """
+        if self.splits and split is None:
+            raise ValueError(
+                "The training dataset has splits, please specify the split you want to read"
+            )
+
         return self._training_dataset_engine.read(self, split, read_options)
 
     def compute_statistics(self):
@@ -248,7 +261,10 @@ class TrainingDataset:
         feature store.
         """
         if self.statistics_config.enabled and engine.get_type() == "spark":
-            return self._statistics_engine.compute_statistics(self, self.read())
+            if self.splits:
+                return self._statistics_engine.register_split_statistics(self)
+            else:
+                return self._statistics_engine.compute_statistics(self, self.read())
 
     def tf_data(
         self,
@@ -258,6 +274,8 @@ class TrainingDataset:
         var_len_features: Optional[list] = [],
         is_training: Optional[bool] = True,
         cycle_length: Optional[int] = 2,
+        deterministic: Optional[bool] = False,
+        file_pattern: Optional[str] = "*.tfrecord*",
     ):
         """
         Returns an object with utility methods to read training dataset as `tf.data.Dataset` object and handle it for further processing.
@@ -267,9 +285,13 @@ class TrainingDataset:
             split: Name of training dataset split. For example, `"train"`, `"test"` or `"val"`, defaults to `None`,
                 returning the full training dataset.
             feature_names: Names of training variables, defaults to `None`.
-            var_len_features: Feature names that have variable length and need to be returned as `tf.io.VarLenFeature`, defaults to `[]`.
-            is_training: Whether it is for training, testing or validation. Defaults to `True`.
+            var_len_features: Feature names that have variable length and need to be returned as `tf.io.VarLenFeature`,
+            defaults to `[]`. is_training: Whether it is for training, testing or validation. Defaults to `True`.
             cycle_length: Number of files to be read and deserialized in parallel, defaults to `2`.
+            deterministic: Controls the order in which the transformation produces elements. If set to False, the
+            transformation is allowed to yield elements out of order to trade determinism for performance.
+            Defaults to `False`.
+            file_pattern: Returns a list of files that match the given pattern Defaults to `*.tfrecord*`.
 
         # Returns
             `TFDataEngine`. An object with utility methods to generate and handle `tf.data.Dataset` object.
@@ -282,6 +304,8 @@ class TrainingDataset:
             var_len_features=var_len_features,
             is_training=is_training,
             cycle_length=cycle_length,
+            deterministic=deterministic,
+            file_pattern=file_pattern,
         )
 
     def show(self, n: int, split: str = None):
@@ -383,8 +407,9 @@ class TrainingDataset:
     @classmethod
     def from_response_json(cls, json_dict):
         json_decamelized = humps.decamelize(json_dict)
-        _ = json_decamelized.pop("type")
-        return cls(**json_decamelized)
+        for td in json_decamelized:
+            _ = td.pop("type")
+        return [cls(**td) for td in json_decamelized]
 
     def update_from_response_json(self, json_dict):
         json_decamelized = humps.decamelize(json_dict)
@@ -666,13 +691,13 @@ class TrainingDataset:
         return self._feature_store_id
 
     @property
-    def prepared_statement_connection(self):
-        """JDBC connection to online features store."""
-        return self._prepared_statement_connection
+    def prepared_statement_engine(self):
+        """JDBC connection engine to retrieve connections to online features store from."""
+        return self._prepared_statement_engine
 
-    @prepared_statement_connection.setter
-    def prepared_statement_connection(self, prepared_statement_connection):
-        self._prepared_statement_connection = prepared_statement_connection
+    @prepared_statement_engine.setter
+    def prepared_statement_engine(self, prepared_statement_engine):
+        self._prepared_statement_engine = prepared_statement_engine
 
     @property
     def prepared_statements(self):
